#!/usr/bin/env python
"""

    run_dless2_rebecca.py
    [--log_file PATH]
    [--verbose]

"""

################################################################################
#
#   test
#
#
#   Copyright (c) 7/13/2010 Rebecca Chodroff
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#################################################################################

import sys, os, re, shutil

# add self to search path for testing
if __name__ == '__main__':
    exe_path = os.path.split(os.path.abspath(sys.argv[0]))[0]
    module_name = os.path.split(sys.argv[0])[1]
    module_name = os.path.splitext(module_name)[0];
else:
    module_name = __name__

# Use import path from <<../python_modules>>
if __name__ == '__main__':
    sys.path.append(os.path.abspath(os.path.join(exe_path,"..", "python_modules")))
    sys.path.insert(0, "/net/cpp-group/Leo/inprogress/oss_projects/ruffus/installation")




#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   options


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


if __name__ == '__main__':
    from optparse import OptionParser
    import StringIO

    parser = OptionParser(version="%prog 1.0", usage = "\n\n    %progs [options]")
    parser.add_option("--targets", dest="targets",
                      metavar="INTEGERS",
                      type="string",
                      help="List of comma separated targets.")
    parser.add_option("--working_dir", dest="working_dir",
                      metavar="PATH",
                      type="string",
                      help="Working directory.")
    parser.add_option("--starting_dir", dest="starting_dir",
                      metavar="PATH",
                      type="string",
                      help="Starting directory.")


    #
    #   general options: verbosity / logging
    #
    parser.add_option("-v", "--verbose", dest = "verbose",
                      action="count", default=0,
                      help="Print more verbose messages for each additional verbose level.")
    parser.add_option("-L", "--log_file", dest="log_file",
                      metavar="FILE",
                      type="string",
                      help="Name and path of log file")
    parser.add_option("--skip_parameter_logging", dest="skip_parameter_logging",
                        action="store_true", default=False,
                        help="Do not print program parameters to log.")
    parser.add_option("--debug", dest="debug",
                        action="count", default=0,
                        help="Set default program parameters in debugging mode.")







    #
    #   pipeline
    #
    parser.add_option("-t", "--target_tasks", dest="target_tasks",
                        action="append",
                        default = list(),
                        metavar="JOBNAME",
                        type="string",
                        help="Target task(s) of pipeline.")
    parser.add_option("-j", "--jobs", dest="jobs",
                        default=1,
                        metavar="N",
                        type="int",
                        help="Allow N jobs (commands) to run simultaneously.")
    parser.add_option("-n", "--just_print", dest="just_print",
                        action="store_true", default=False,
                        help="Don't actually run any commands; just print the pipeline.")
    parser.add_option("--flowchart", dest="flowchart",
                        metavar="FILE",
                        type="string",
                        help="Don't actually run any commands; just print the pipeline "
                             "as a flowchart.")
    parser.add_option("--colour_scheme_index", dest="colour_scheme_index",
                        metavar="INTEGER",
                        type="int",
                        help="Index of colour scheme for flow chart.")

    #
    #   Less common pipeline options
    #
    parser.add_option("--key_legend_in_graph", dest="key_legend_in_graph",
                        action="store_true", default=False,
                        help="Print out legend and key for dependency graph.")
    parser.add_option("--forced_tasks", dest="forced_tasks",
                        action="append",
                        default = list(),
                        metavar="JOBNAME",
                        type="string",
                        help="Pipeline task(s) which will be included even if they are up to date.")

    # get help string
    f =StringIO.StringIO()
    parser.print_help(f)
    helpstr = f.getvalue()
    original_args = " ".join(sys.argv)
    (options, remaining_args) = parser.parse_args()


    #vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
    #                                             #
    #   Debug: Change these                       #
    #                                             #
    #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    if options.debug:
        options.log_file                = os.path.join("run_dless2.log")
        if not options.verbose:
            options.verbose             = 1
        if not options.targets:
            options.targets             = "87"
        if not options.working_dir:
            options.working_dir         = "DLESS"
        if not options.starting_dir:
            options.starting_dir         = "DUMMY"
    #vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
    #                                             #
    #   Debug: Change these                       #
    #                                             #
    #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    #
    #   mandatory options
    #
    mandatory_options = ["targets", "starting_dir", "working_dir"]
    def check_mandatory_options (options, mandatory_options, helpstr):
        """
        Check if specified mandatory options have b een defined
        """
        missing_options = []
        for o in mandatory_options:
            if not getattr(options, o):
                missing_options.append("--" + o)

        if not len(missing_options):
            return

        raise Exception("Missing mandatory parameter%s: %s.\n\n%s\n\n" %
                        ("s" if len(missing_options) > 1 else "",
                         ", ".join(missing_options),
                         helpstr))
    check_mandatory_options (options, mandatory_options, helpstr)




#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

from ruffus import *
from ruffus.ruffus_exceptions import JobSignalledBreak

#from json import dumps
#from collections import defaultdict


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Constants


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#from read_project_parameter_file import get_keys
#parameters =  get_keys('PARAMETERS')
#
#reference_species = parameters['REFERENCE_SPECIES']
#species = parameters['SPECIES']
#working_dir = parameters['WORKING_DIR_ROOT']
#tree = parameters['TREE']
#ref_sequences = parameters['REFERENCE_SEQUENCES']
#working_dir= parameters['WORKING_DIR_ROOT']
#tba_alignments = parameters['TBA_ALIGNMENTS_DIR']
#tba_projected_alignments = parameters['TBA_PROJECTED_ALIGNMENTS_DIR']
#fasta_alignments = parameters['FASTA_ALIGNMENTS_DIR']
#repeats_dir = parameters['REPEATS_DIR']
#neutral_mods_dir = parameters['NEUTRAL_MODELS_DIR']
#indel_hist_dir = parameters['INDEL_HISTORY_DIR']
#indel_mods_dir = parameters['INDEL_MODELS_DIR']
#
#python_code = parameters['PYTHON_CODE_DIR']
#find_ars = parameters['FIND_ARS']
#maf_project = parameters['MAF_PROJECT_BIN']
#phylo_fit = parameters['PHYLOFIT_BINARY']
#msa_view = parameters['MSA_VIEW_BINARY']
#indel_history = parameters['INDEL_HISTORY_BINARY']
#tree_doctor = parameters['TREE_DOCTOR_BINARY']
#indel_fit = parameters['INDEL_FIT_BINARY']
#dless = parameters['DLESS_BINARY']


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Functions


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


#def get_all_alignment_file_names():
#    alignment_file_names = os.listdir(tba_alignments)
#    return alignment_file_names
#
#def get_all_targets():
#    alignment_file_names = get_all_alignment_file_names()
#    targets = []
#    for alignment_file_name in alignment_file_names:
#        targets.append(alignment_file_name.split('.')[0])
#    return targets
#
#
#def find_ARs(target):
#    coor_file = os.path.join(ref_sequences, target + '_' + reference_species)
#    coor = open(coor_file, 'r').readline()
#    species, chr, start, strand, other = coor.split(':',4)
#    RM_out = os.path.join(repeats_dir, target + '_' + reference_species + '.out')
#    out_file = os.path.join(repeats_dir, target + '_' + reference_species + '.ar')
#    os.system('perl ' + find_ars + ' ' + RM_out + ' >' + out_file)
#    return chr, start, strand, out_file
#
#def write_gff(target):
#    chr, start, strand, out_file = find_ARs(target)
#    file = open(out_file, 'r')
#    line = file.readline()
#    lines = file.readlines()
#    repeats_file = os.path.join(repeats_dir, target + '_' + reference_species + '.gff')
#    repeats = open(repeats_file, 'w')
#    for line in lines:
#        line = line.strip()
#        beg, end, feature = line.split()
#        b = str(int(beg) + int(start))
#        e = str(int(end) + int(start))
#        entry = '\t'.join([chr, 'RepeatMasker', 'AR', b, e,'.', '.','.',feature])
#        repeats.write(entry + '\n')
#    repeats.close()
#    return repeats_file




#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Logger


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

if __name__ == '__main__':
    import logging
    import logging.handlers

    MESSAGE = 15
    logging.addLevelName(MESSAGE, "MESSAGE")

    def setup_std_logging (logger, log_file, verbose):
        """
        set up logging using programme options
        """
        class debug_filter(logging.Filter):
            """
            Ignore INFO messages
            """
            def filter(self, record):
                return logging.INFO != record.levelno

        class NullHandler(logging.Handler):
            """
            for when there is no logging
            """
            def emit(self, record):
                pass

        # We are interesting in all messages
        logger.setLevel(logging.DEBUG)
        has_handler = False

        # log to file if that is specified
        if log_file:
            handler = logging.FileHandler(log_file, delay=False)
            handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)6s - %(message)s"))
            handler.setLevel(MESSAGE)
            logger.addHandler(handler)
            has_handler = True

        # log to stderr if verbose
        if verbose:
            stderrhandler = logging.StreamHandler(sys.stderr)
            stderrhandler.setFormatter(logging.Formatter("    %(message)s"))
            stderrhandler.setLevel(logging.DEBUG)
            if log_file:
                stderrhandler.addFilter(debug_filter())
            logger.addHandler(stderrhandler)
            has_handler = True

        # no logging
        if not has_handler:
            logger.addHandler(NullHandler())


    #
    #   set up log
    #
    logger = logging.getLogger(module_name)
    setup_std_logging(logger, options.log_file, options.verbose)

    #
    #   Allow logging across Ruffus pipeline
    #
    def get_logger (logger_name, args):
        return logger

    from ruffus.proxy_logger import *
    (logger_proxy,
     logging_mutex) = make_shared_logger_and_proxy (get_logger,
                                                    module_name,
                                                    {})

    #
    #   log programme parameters
    #
    if not options.skip_parameter_logging:
        programme_name = os.path.split(sys.argv[0])[1]
        logger.info("%s %s" % (programme_name, original_args))


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Pipeline


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#
#   convert targets into individual strings
#       and identify files in staring_dir
#
options.targets = re.split(", *", options.targets)
if not len(options.targets):
    raise Exception ("Please specify the targets as a common separated list for --targets")
starting_files = [os.path.join(options.starting_dir, t + ".maf") for t in options.targets]
repeats_files = [os.path.join(options.starting_dir, t + ".repeats") for t in options.targets]

#
# regex to split path out and substitute working directory
#
@follows(mkdir(options.working_dir, "test"))
@transform(starting_files, regex(r"(.+/)([^/]+\.maf)"), os.path.join(options.working_dir, r"\2"))
def copy_maf_into_working_directory(input_file, output_file):
    """
    Make copy in working directory
    """
    shutil.copyfile(input_file, output_file)

@follows(mkdir(options.working_dir))
@transform(repeats_files, regex(r"(.+/)([^/]+\.repeats)"), os.path.join(options.working_dir, r"\2"))
def copy_repeats_into_working_directory(input_file, output_file):
    """
    Make copy in working directory
    """
    shutil.copyfile(input_file, output_file)


#
#   pipes in output from copy_maf_into_working_directory
#
#      working_dir/target.maf ->   working_dir/target.projected_maf
#
@transform(copy_maf_into_working_directory, suffix(".maf"), ".projected_maf")
def project_maf_alignments(input_file, output_file):
    #os.system(maf_project + ' ' + input_file + ' ' + reference_species + ' > ' + output_file)
    open(output_file, "w")

@transform(project_maf_alignments, suffix(".projected_maf"), add_inputs(r"\1.repeats"), r".neutral_model")
@follows(copy_repeats_into_working_directory)
def generate_neutral_model(input_files, output_file):
    maf_file, repeats_file = input_files
    #cmd_str = (phylo_fit + ' --tree ' + tree
    #              + ' --features ' + repeats_file
    #              + ' --do-cats AR --out-root '
    #              + output_file + ' --msa-format MAF '
    #              + maf_file)
    #os.system(cmd_str)
    #run_cmd(cmd_str, "generate neutral model")
    #queue_cmd_prefix    = "qrsh -now n -cwd -p -6 -v BASH_ENV=~/.bashrc -q medium_jobs.q"
    #target_name            = os.path.splitext("maf_file")[0]
    #run_cmd(cmd_str, "generate neutral model", queue_cmd_prefix = queue_cmd_prefix, job_name = target_name)
    open(output_file, "w")


#must convert maf to fasta for indel history program
@transform(project_maf_alignments, suffix(".projected_maf"), ".fa")
def convert_maf2fasta(input_file, output_file):
    #species = (open(species, 'r')).readline()
    #os.system(msa_view + ' ' + input_file\
    #          + ' --soft-masked --seqs '\
    #          + species + ' --in-format MAF > ' + output_file)
    open(output_file, "w")


@follows(generate_neutral_model)
@transform(convert_maf2fasta, regex("(.+).fa"), add_inputs(r"\1.neutral_model"), r"\1.indel_history")
def generate_indel_history(input_files, output_file):
    fasta_file, neutral_model = input_files
    #os.system(indel_history + ' ' + fasta_file + ' ' + neutral_model + ' > ' + output_file)
    open(output_file, "w")

def parameters_for_dless(indel_history_file, neutral_model):
    os.system(tree_doctor + ' -t ' + neutral_model + ' > ' + tree_file)
    cmd = indel_fit + ' ' + indel_history_file + ' ' + tree_file
    fin,fout=os.popen4(cmd)
    indel_mod = fout.read()
    indelmod=[]
    for n in [2,5,8]:
        indelmod.append(indel_mod.split()[n].strip(','))
    return indelmod


@follows(generate_neutral_model)
@follows(convert_maf2fasta)
@transform(generate_indel_history,
            suffix(".indel_history"),
           add_inputs(r"\1.neutral_model", r"\1.fa"),
           ".dless.out")
def run_dless(input_files, output_file):
    indel_history_file, neutral_model, fasta_file = input_files
    #indelmod = parameters_for_dless(indel_history_file, neutral_model)
    #cmd = ' '.join([dless,'-I',','.join(indel_params), '-H', indel_history_file,
    #                alignment, neutral_mod, '>', output_file])
    #os.system(cmd)
    open(output_file, "w")











#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
if __name__ == '__main__':
    if options.just_print:
        pipeline_printout(sys.stdout, options.target_tasks, options.forced_tasks,
                            verbose=options.verbose)

    elif options.flowchart:
        pipeline_printout_graph (   open(options.flowchart, "w"),
                                    os.path.splitext(options.flowchart)[1][1:],
                                    options.target_tasks,
                                    options.forced_tasks,
                                    no_key_legend   = not options.key_legend_in_graph,
                                    minimal_key_legend             = True,
                                    user_colour_scheme = {"colour_scheme_index": options.colour_scheme_index},
                                    pipeline_name = "dless2")
        #graph_colour_demo_printout (open(options.flowchart, "w"),
    #                                os.path.splitext(options.flowchart)[1][1:])
    else:
        pipeline_run(options.target_tasks, options.forced_tasks,
                            multiprocess    = options.jobs,
                            logger          = stderr_logger,
                            verbose         = options.verbose)











#! /usr/bin/env python

from __future__ import print_function
import unittest
import glob
from ruffus import pipeline_run, Pipeline, follows, posttask, collate, mkdir, regex, files
import os
import sys
if sys.hexversion < 0x03000000:
    from future_builtins import zip
"""

    test_N_x_M_and_collate.py


        This script takes N pairs of input file pairs
                                (with the suffices .gene and .gwas)
                          and runs them against M sets of simulation data
                                (with the suffix .simulation)
        A summary per input file pair is then produced


    In pseudo-code:

        STEP_1:

        for n_file in NNN_pairs_of_input_files:
            for m_file in MMM_simulation_data:

                [n_file.gene,
                 n_file.gwas,
                 m_file.simulation] -> n_file.m_file.simulation_res


        STEP_2:

        for n_file in NNN_pairs_of_input_files:

            n_file.*.simulation_res -> n_file.mean


        n = CNT_GENE_GWAS_FILES
        m = CNT_SIMULATION_FILES



"""

CNT_GENE_GWAS_FILES = 2
CNT_SIMULATION_FILES = 3


# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0]))
gene_data_dir = "%s/temp_gene_data_for_intermediate_example" % tempdir
simulation_data_dir = "%s/temp_simulation_data_for_intermediate_example" % tempdir
working_dir = "%s/working_dir_for_intermediate_example" % tempdir


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

try:
    import StringIO as io
except:
    import io as io

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Functions


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

# _________________________________________________________________________________________
#
#   get gene gwas file pairs
#
# _________________________________________________________________________________________
def get_gene_gwas_file_pairs():
    """
    Helper function to get all *.gene, *.gwas from the direction specified
        in --gene_data_dir

    Returns
        file pairs with both .gene and .gwas extensions,
        corresponding roots (no extension) of each file
    """

    gene_files = glob.glob(os.path.join(gene_data_dir, "*.gene"))
    gwas_files = glob.glob(os.path.join(gene_data_dir, "*.gwas"))

    common_roots = set([os.path.splitext(os.path.split(x)[1])[0]
                        for x in gene_files])
    common_roots &= set([os.path.splitext(os.path.split(x)[1])[0]
                         for x in gwas_files])
    common_roots = list(common_roots)

    p = os.path
    g_dir = gene_data_dir

    file_pairs = [[p.join(g_dir, x + ".gene"), p.join(g_dir, x + ".gwas")]
                  for x in common_roots]

    return file_pairs, common_roots

# _________________________________________________________________________________________
#
#   get simulation files
#
# _________________________________________________________________________________________


def get_simulation_files():
    """
    Helper function to get all *.simulation from the direction specified
        in --simulation_data_dir
        Returns
            file with .simulation extensions,
            corresponding roots (no extension) of each file
    """
    simulation_files = glob.glob(os.path.join(
        simulation_data_dir, "*.simulation"))
    simulation_roots = [os.path.splitext(os.path.split(x)[1])[
        0] for x in simulation_files]
    return simulation_files, simulation_roots


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# _________________________________________________________________________________________
#
#   setup_simulation_data
#
# _________________________________________________________________________________________

#
# mkdir: makes sure output directories exist before task
#
@follows(mkdir(gene_data_dir, simulation_data_dir))
def setup_simulation_data():
    """
    create simulation files
    """
    for i in range(CNT_GENE_GWAS_FILES):
        open(os.path.join(gene_data_dir, "%03d.gene" % i), "w").close()
        open(os.path.join(gene_data_dir, "%03d.gwas" % i), "w").close()

    # gene files without corresponding gwas and vice versa
    open(os.path.join(gene_data_dir, "orphan1.gene"), "w").close()
    open(os.path.join(gene_data_dir, "orphan2.gwas"), "w").close()
    open(os.path.join(gene_data_dir, "orphan3.gwas"), "w").close()

    for i in range(CNT_SIMULATION_FILES):
        open(os.path.join(simulation_data_dir, "%03d.simulation" % i), "w").close()


# _________________________________________________________________________________________
#
#   cleanup_simulation_data
#
# _________________________________________________________________________________________
def try_rmdir(d):
    if os.path.exists(d):
        try:
            os.rmdir(d)
        except OSError:
            sys.stderr.write(
                "Warning:\t%s is not empty and will not be removed.\n" % d)


def cleanup_simulation_data():
    """
    cleanup files
    """

    #
    #   cleanup gene and gwas files
    #
    for f in glob.glob(os.path.join(gene_data_dir, "*.gene")):
        os.unlink(f)
    for f in glob.glob(os.path.join(gene_data_dir, "*.gwas")):
        os.unlink(f)
    try_rmdir(gene_data_dir)

    #
    #   cleanup simulation
    #
    for f in glob.glob(os.path.join(simulation_data_dir, "*.simulation")):
        os.unlink(f)
    try_rmdir(simulation_data_dir)

    #
    #   cleanup working_dir
    #
    for f in glob.glob(os.path.join(working_dir, "simulation_results", "*.simulation_res")):
        os.unlink(f)
    try_rmdir(os.path.join(working_dir, "simulation_results"))

    for f in glob.glob(os.path.join(working_dir, "*.mean")):
        os.unlink(f)
    try_rmdir(working_dir)

    try_rmdir(tempdir)


# _________________________________________________________________________________________
#
#   Step 1:
#
#        for n_file in NNN_pairs_of_input_files:
#            for m_file in MMM_simulation_data:
#
#                [n_file.gene,
#                 n_file.gwas,
#                 m_file.simulation] -> working_dir/n_file.m_file.simulation_res
#
# _________________________________________________________________________________________
def generate_simulation_params():
    """
    Custom function to generate
    file names for gene/gwas simulation study
    """

    simulation_files, simulation_file_roots = get_simulation_files()
    gene_gwas_file_pairs, gene_gwas_file_roots = get_gene_gwas_file_pairs()

    for sim_file, sim_file_root in zip(simulation_files, simulation_file_roots):
        for (gene, gwas), gene_file_root in zip(gene_gwas_file_pairs, gene_gwas_file_roots):

            result_file = "%s.%s.simulation_res" % (
                gene_file_root, sim_file_root)
            result_file_path = os.path.join(
                working_dir, "simulation_results", result_file)

            yield [gene, gwas, sim_file], result_file_path, gene_file_root, sim_file_root, result_file

#
# mkdir: makes sure output directories exist before task
#


@follows(setup_simulation_data)
@follows(mkdir(working_dir, os.path.join(working_dir, "simulation_results")))
@files(generate_simulation_params)
def gwas_simulation(input_files, result_file_path, gene_file_root, sim_file_root, result_file):
    """
    Dummy calculation of gene gwas vs simulation data
    Normally runs in parallel on a computational cluster
    """
    (gene_file,
     gwas_file,
     simulation_data_file) = input_files

    simulation_res_file = open(result_file_path, "w")
    simulation_res_file.write("%s + %s -> %s\n" %
                              (gene_file_root, sim_file_root, result_file))
    simulation_res_file.close()


# _________________________________________________________________________________________
#
#   Step 2:
#
#       Statistical summary per gene/gwas file pair
#
#        for n_file in NNN_pairs_of_input_files:
#            working_dir/simulation_results/n.*.simulation_res
#               -> working_dir/n.mean
#
# _________________________________________________________________________________________


@collate(gwas_simulation, regex(r"simulation_results/(\d+).\d+.simulation_res"), r"\1.mean")
@posttask(lambda: sys.stdout.write("\nOK\n"))
def statistical_summary(result_files, summary_file):
    """
    Simulate statistical summary
    """

    summary_file = open(summary_file, "w")
    for f in result_files:
        with open(f) as ii:
            summary_file.write(ii.read())
    summary_file.close()


try:
    from StringIO import StringIO
except:
    from io import StringIO


class Test_ruffus(unittest.TestCase):

    def tearDown(self):
        try:
            cleanup_simulation_data()
            pass
        except:
            pass

    def test_ruffus(self):
        pipeline_run(multiprocess=50, verbose=0, pipeline="main")
        for oo in "000.mean", "001.mean":
            results_file_name = os.path.join(working_dir, oo)
            if not os.path.exists(results_file_name):
                raise Exception("Missing %s" % results_file_name)

    def test_newstyle_ruffus(self):

        test_pipeline = Pipeline("test")

        test_pipeline.follows(setup_simulation_data, mkdir(
            gene_data_dir, simulation_data_dir))

        test_pipeline.files(gwas_simulation, generate_simulation_params)\
            .follows(setup_simulation_data)\
            .follows(mkdir(working_dir, os.path.join(working_dir, "simulation_results")))

        test_pipeline.collate(statistical_summary, gwas_simulation, regex(r"simulation_results/(\d+).\d+.simulation_res"), r"\1.mean")\
            .posttask(lambda: sys.stdout.write("\nOK\n"))

        test_pipeline.run(multiprocess=50, verbose=0)
        for oo in "000.mean", "001.mean":
            results_file_name = os.path.join(working_dir, oo)
            if not os.path.exists(results_file_name):
                raise Exception("Missing %s" % results_file_name)


if __name__ == '__main__':
    unittest.main()



################################################################################
#
#
#   cmd_line_helper.py
#
#   Copyright (c) 10/9/2009 Leo Goodstadt
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#################################################################################
import sys
from .ruffus_utility import CHECKSUM_REGENERATE
from . import proxy_logger
from . import task
import logging.handlers
import logging
"""

********************************************
:mod:`ruffus.cmdline` -- Overview
********************************************

.. moduleauthor:: Leo Goodstadt <ruffus@llew.org.uk>


    #
    #   Using argparse (new in python v 2.7)
    #
    from ruffus import *

    parser = cmdline.get_argparse(   description='WHAT DOES THIS PIPELINE DO?')

    parser.add_argument("--input_file")

    options = parser.parse_args()

    #  logger which can be passed to ruffus tasks
    logger, logger_mutex = cmdline.setup_logging (__name__, options.log_file, options.verbose)

    #_____________________________________________________________________________________

    #   pipelined functions go here

    #_____________________________________________________________________________________

    cmdline.run (options)



    #
    #   Using optparse (new in python v 2.6)
    #
    from ruffus import *

    parser = cmdline.get_optgparse(version="%prog 1.0", usage = "\n\n    %prog [options]")

    parser.add_option("-c", "--custom", dest="custom", action="count")


    (options, remaining_args) = parser.parse_args()

    #  logger which can be passed to ruffus tasks
    logger, logger_mutex = cmdline.setup_logging ("this_program", options.log_file, options.verbose)

    #_____________________________________________________________________________________

    #   pipelined functions go here

    #_____________________________________________________________________________________

    cmdline.run (options)


"""


#
#   print options
#
flowchart_formats = ["svg", "svgz", "png",
                     "jpg", "psd", "tif", "eps", "pdf", "dot"]
#  "jpeg", "gif", "plain", "ps", "wbmp", "canon",
#  "cmap", "cmapx", "cmapx_np", "fig", "gd", "gd2",
# "gv", "imap", "imap_np", "ismap", "jpe", "plain-ext",
# "ps2", "tk", "vml", "vmlz", "vrml", "x11", "xdot", "xlib"
# Replace last comma with " and". Mad funky, unreadable reverse replace code: couldn't resist!
flowchart_formats_str = ", ".join(["%r" % ss for ss in flowchart_formats])[
    ::-1].replace(" ,", ", or "[::-1], 1)[::-1]


# _________________________________________________________________________________________

#   get_argparse
# _________________________________________________________________________________________
def get_argparse(*args, **args_dict):
    """
    Set up argparse
        to allow for ruffus specific options:

            --verbose
            --version
            --log_file

        -t, --target_tasks
        -j, --jobs
        -n, --just_print
            --flowchart
            --touch_files_only
            --recreate_database
            --checksum_file_name
            --key_legend_in_graph
            --draw_graph_horizontally
            --flowchart_format
            --forced_tasks

    Optionally specify ignored_args = ["verbose", "recreate_database",...]
        listing names which will not be added as valid options on the command line

    Optionally specify version = "%(prog)s version 1.234"

    """
    import argparse

    # version and ignored_args are for append_to_argparse
    orig_args_dict = dict(args_dict)
    if "version" in args_dict:
        del args_dict["version"]
    if "ignored_args" in args_dict:
        del args_dict["ignored_args"]

    parser = argparse.ArgumentParser(*args, **args_dict)

    return append_to_argparse(parser, **orig_args_dict)

# _________________________________________________________________________________________

#   append_to_argparse
# _________________________________________________________________________________________


def append_to_argparse(parser, **args_dict):
    """
    Common options:

            --verbose
            --version
            --log_file
    """

    if "version" in args_dict:
        prog_version = args_dict["version"]
    else:
        prog_version = "%(prog)s 1.0"

    #
    #   ignored_args contains a list of options which will *not* be added
    #
    if "ignored_args" in args_dict:
        if isinstance(args_dict["ignored_args"], str):
            ignored_args = set([args_dict["ignored_args"]])
        else:
            try:
                ignored_args = set(args_dict["ignored_args"])
            except:
                raise Exception(
                    "Error: expected ignored_args = ['list_of', 'option_names']")
    else:
        ignored_args = set()

    common_options = parser.add_argument_group('Common options')
    if "verbose" not in ignored_args:
        common_options.add_argument(
            '--verbose', "-v", const="+", default=[], nargs='?',
            action="append",
            help="Print more verbose messages for each additional verbose level.")
    if "version" not in ignored_args:
        common_options.add_argument(
            '--version', action='version', version=prog_version)
    if "log_file" not in ignored_args:
        common_options.add_argument("-L", "--log_file", metavar="FILE", type=str,
                                    help="Name and path of log file")

    #
    #   pipeline
    #
    pipeline_options = parser.add_argument_group('pipeline arguments')
    if "target_tasks" not in ignored_args:
        pipeline_options.add_argument("-T", "--target_tasks", action="append",
                                      metavar="JOBNAME", type=str,
                                      help="Target task(s) of pipeline.", default=[])
    if "jobs" not in ignored_args:
        pipeline_options.add_argument("-j", "--jobs", default=1, metavar="N", type=int,
                                      help="Allow N jobs (commands) to run simultaneously.")
    if "use_threads" not in ignored_args:
        pipeline_options.add_argument("--use_threads", action="store_true",
                                      help="Use multiple threads rather than processes. Needs --jobs N with N > 1")
    if "just_print" not in ignored_args:
        pipeline_options.add_argument("-n", "--just_print", action="store_true",
                                      help="Don't actually run any commands; just print the pipeline.")
    if "touch_files_only" not in ignored_args:
        pipeline_options.add_argument("--touch_files_only", action="store_true",
                                      help="Don't actually run any commands; just 'touch' the output for each task to make them appear up to date.")
    if "recreate_database" not in ignored_args:
        pipeline_options.add_argument("--recreate_database", action="store_true",
                                      help="Don't actually run any commands; just recreate the checksum database.")
    if "checksum_file_name" not in ignored_args:
        pipeline_options.add_argument("--checksum_file_name", dest="history_file", metavar="FILE", type=str,
                                      help="Path of the checksum file.")
    if "flowchart" not in ignored_args:
        pipeline_options.add_argument("--flowchart", metavar="FILE", type=str,
                                      help="Don't run any commands; just print pipeline as a flowchart.")

    #
    #   Less common pipeline options
    #
    if "key_legend_in_graph" not in ignored_args:
        pipeline_options.add_argument("--key_legend_in_graph",     action="store_true",
                                      help="Print out legend and key for dependency graph.")
    if "draw_graph_horizontally" not in ignored_args:
        pipeline_options.add_argument("--draw_graph_horizontally", action="store_true", dest="draw_horizontally",
                                      help="Draw horizontal dependency graph.")
    if "flowchart_format" not in ignored_args:
        pipeline_options.add_argument("--flowchart_format", metavar="FORMAT",
                                      type=str, choices=flowchart_formats,
                                      default=None,
                                      help="format of dependency graph file. Can be %s. Defaults to the "
                                      "file name extension of --flowchart FILE." % flowchart_formats_str)
    if "forced_tasks" not in ignored_args:
        pipeline_options.add_argument("--forced_tasks", action="append",
                                      metavar="JOBNAME", type=str,
                                      help="Task(s) which will be included even if they are up to date.", default=[])

    return parser


# _________________________________________________________________________________________

#   Hacky extension to *deprecated!!* optparse to support variable number of arguments
#       for --verbose
# _________________________________________________________________________________________
def vararg_callback(option, opt_str, value, parser):
    #
    #   get current value
    #
    if hasattr(parser.values, option.dest):
        value = getattr(parser.values, option.dest)
    else:
        value = []
    if not len(parser.rargs):
        value.append("+")
    else:
        arg = parser.rargs[0]
        # stop on --foo like options
        if arg[:1] == "-":
            value.append("+")
        else:
            value.append(arg)
            del parser.rargs[:1]
    setattr(parser.values, option.dest, value)


# _________________________________________________________________________________________

#   optparse is deprecated!
# _________________________________________________________________________________________
def get_optparse(*args, **args_dict):
    """
    Set up OptionParser from optparse
        to allow for ruffus specific options:

            --verbose
            --version
            --log_file

        -t, --target_tasks
        -j, --jobs
        -n, --just_print
            --flowchart
            --touch_files_only
            --recreate_database
            --checksum_file_name
            --key_legend_in_graph
            --draw_graph_horizontally
            --flowchart_format
            --forced_tasks

    Optionally specify ignored_args = ["verbose", "recreate_database",...]
        listing names which will not be added as valid options on the command line

    N.B. optparse is deprecated since python version 2.7.
    """
    from optparse import OptionParser

    # ignored_args are for append_to_optparse
    orig_args_dict = dict(args_dict)
    if "ignored_args" in args_dict:
        del args_dict["ignored_args"]

    parser = OptionParser(*args, **args_dict)

    return append_to_optparse(parser, **orig_args_dict)

# _________________________________________________________________________________________

#   optparse is deprecated!
# _________________________________________________________________________________________


def append_to_optparse(parser, **args_dict):
    """
    Set up OptionParser from optparse
        to allow for ruffus specific options:

            --verbose
            --version
            --log_file

        -t, --target_tasks
        -j, --jobs
        -n, --just_print
            --flowchart
            --touch_files_only
            --recreate_database
            --checksum_file_name
            --key_legend_in_graph
            --draw_graph_horizontally
            --flowchart_format
            --forced_tasks

    Optionally specify ignored_args = ["verbose", "recreate_database",...]
        listing names which will not be added as valid options on the command line
    """

    #
    #   ignored_args contains a list of options which will *not* be added
    #
    if "ignored_args" in args_dict:
        if isinstance(args_dict["ignored_args"], str):
            ignored_args = set([args_dict["ignored_args"]])
        else:
            try:
                ignored_args = set(args_dict["ignored_args"])
            except:
                raise Exception(
                    "Error: expected ignored_args = ['list_of', 'option_names']")
    else:
        ignored_args = set()

    #
    #   general options: verbosity / logging
    #
    if "verbose" not in ignored_args:
        parser.add_option("-v", "--verbose", dest="verbose",
                          action="callback", default=[],
                          # ---------------------------------------------------------------
                          # hack to get around unreasonable discrimination against
                          # --long_options=with_equals in opt_parse::_process_long_opt()
                          # when using a callback
                          type=int,
                          nargs=0,
                          # ---------------------------------------------------------------
                          callback=vararg_callback,
                          help="Print more verbose messages for each additional verbose level.")
    if "log_file" not in ignored_args:
        parser.add_option("-L", "--log_file", dest="log_file",
                          metavar="FILE",
                          type="string",
                          help="Name and path of log file")
    #
    #   pipeline
    #
    if "target_tasks" not in ignored_args:
        parser.add_option("-t", "--target_tasks", dest="target_tasks",
                          action="append",
                          default=list(),
                          metavar="JOBNAME",
                          type="string",
                          help="Target task(s) of pipeline.")
    if "jobs" not in ignored_args:
        parser.add_option("-j", "--jobs", dest="jobs",
                          default=1,
                          metavar="N",
                          type="int",
                          help="Allow N jobs (commands) to run simultaneously.")
    if "use_threads" not in ignored_args:
        parser.add_option("--use_threads", dest="use_threads",
                          action="store_true", default=False,
                          help="Use multiple threads rather than processes. Needs --jobs N with N > 1")
    if "just_print" not in ignored_args:
        parser.add_option("-n", "--just_print", dest="just_print",
                          action="store_true", default=False,
                          help="Don't actually run any commands; just print the pipeline.")
    if "touch_files_only" not in ignored_args:
        parser.add_option("--touch_files_only", dest="touch_files_only",
                          action="store_true", default=False,
                          help="Don't actually run any commands; just 'touch' the output for each task to make them appear up to date.")
    if "recreate_database" not in ignored_args:
        parser.add_option("--recreate_database", dest="recreate_database",
                          action="store_true", default=False,
                          help="Don't actually run any commands; just recreate the checksum database.")
    if "checksum_file_name" not in ignored_args:
        parser.add_option("--checksum_file_name", dest="history_file",
                          metavar="FILE",
                          type="string",
                          help="Path of the checksum file.")
    if "flowchart" not in ignored_args:
        parser.add_option("--flowchart", dest="flowchart",
                          metavar="FILE",
                          type="string",
                          help="Don't actually run any commands; just print the pipeline "
                          "as a flowchart.")

    #
    #   Less common pipeline options
    #
    if "key_legend_in_graph" not in ignored_args:
        parser.add_option("--key_legend_in_graph", dest="key_legend_in_graph",
                          action="store_true", default=False,
                          help="Print out legend and key for dependency graph.")
    if "flowchart_format" not in ignored_args:
        parser.add_option("--draw_graph_horizontally", dest="draw_horizontally",
                          action="store_true", default=False,
                          help="Draw horizontal dependency graph.")
    if "flowchart_format" not in ignored_args:
        parser.add_option("--flowchart_format", dest="flowchart_format",
                          metavar="FORMAT",
                          type="string",
                          default=None,
                          help="format of dependency graph file. Can be %s. Defaults to the "
                          "file name extension of --flowchart FILE." % flowchart_formats_str)
    if "forced_tasks" not in ignored_args:
        parser.add_option("--forced_tasks", dest="forced_tasks",
                          action="append",
                          default=list(),
                          metavar="JOBNAME",
                          type="string",
                          help="Pipeline task(s) which will be included even if they are up to date.")

        return parser


MESSAGE = 15
logging.addLevelName(MESSAGE, "MESSAGE")

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Logger


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#   Allow logging across Ruffus pipeline
#
def setup_logging(module_name, log_file_name, verbose):
    return proxy_logger.make_shared_logger_and_proxy(setup_logging_factory, module_name, [log_file_name, verbose])


def setup_logging_factory(logger_name, args):
    log_file_name, verbose = args
    """
    This function is a simple around wrapper around the python
    `logging <http://docs.python.org/library/logging.html>`_ module.

    This *logger_factory* example creates logging objects which can
    then be managed by proxy via ``ruffus.proxy_logger.make_shared_logger_and_proxy()``

    This can be:

        * a `disk log file <http://docs.python.org/library/logging.html#filehandler>`_
        * a automatically backed-up `(rotating) log <http://docs.python.org/library/logging.html#rotatingfilehandler>`_.
        * any log specified in a `configuration file <http://docs.python.org/library/logging.html#configuration-file-format>`_

    These are specified in the ``args`` dictionary forwarded by ``make_shared_logger_and_proxy()``

    :param logger_name: name of log
    :param args: a dictionary of parameters forwarded from ``make_shared_logger_and_proxy()``

        Valid entries include:

            .. describe:: "level"

                Sets the `threshold <http://docs.python.org/library/logging.html#logging.Handler.setLevel>`_ for the logger.

            .. describe:: "config_file"

                The logging object is configured from this `configuration file <http://docs.python.org/library/logging.html#configuration-file-format>`_.

            .. describe:: "file_name"

                Sets disk log file name.

            .. describe:: "rotating"

                Chooses a `(rotating) log <http://docs.python.org/library/logging.html#rotatingfilehandler>`_.

            .. describe:: "maxBytes"

                Allows the file to rollover at a predetermined size

            .. describe:: "backupCount"

                If backupCount is non-zero, the system will save old log files by appending the extensions ``.1``, ``.2``, ``.3`` etc., to the filename.

            .. describe:: "delay"

                Defer file creation until the log is written to.

            .. describe:: "formatter"

                `Converts <http://docs.python.org/library/logging.html#formatter-objects>`_ the message to a logged entry string.
                For example,
                ::

                    "%(asctime)s - %(name)s - %(levelname)6s - %(message)s"



    """

    #
    #   Log file name with logger level
    #
    new_logger = logging.getLogger(logger_name)

    class debug_filter(logging.Filter):
        """
        Ignore INFO messages
        """

        def filter(self, record):
            return logging.INFO != record.levelno

    class NullHandler(logging.Handler):
        """
        for when there is no logging
        """

        def emit(self, record):
            pass

    # We are interesting in all messages
    new_logger.setLevel(logging.DEBUG)
    has_handler = False

    # log to file if that is specified
    if log_file_name:
        handler = logging.FileHandler(log_file_name, delay=False)

        class stipped_down_formatter(logging.Formatter):
            def format(self, record):
                prefix = ""
                if not hasattr(self, "first_used"):
                    self.first_used = True
                    prefix = "\n" + self.formatTime(record, "%Y-%m-%d")
                    prefix += " %(name)s\n" % record.__dict__
                if record.levelname in ("INFO", "MESSAGE", "DEBUG"):
                    self._fmt = " %(asctime)s - %(message)s"
                else:
                    self._fmt = " %(asctime)s - %(levelname)-7s - %(message)s"
                return prefix + logging.Formatter.format(self, record)
        handler.setFormatter(stipped_down_formatter(
            "%(asctime)s - %(name)s - %(levelname)6s - %(message)s", "%H:%M:%S"))
        handler.setLevel(MESSAGE)
        new_logger.addHandler(handler)
        has_handler = True

    # log to stderr if verbose
    if verbose:
        stderrhandler = logging.StreamHandler(sys.stderr)
        stderrhandler.setFormatter(logging.Formatter("    %(message)s"))
        stderrhandler.setLevel(logging.DEBUG)
        if log_file_name:
            stderrhandler.addFilter(debug_filter())
        new_logger.addHandler(stderrhandler)
        has_handler = True

    # no logging
    if not has_handler:
        new_logger.addHandler(NullHandler())

    #
    #   This log object will be wrapped in proxy
    #
    return new_logger


#
#   valid arguments to each function which are not exposed by any options in the command line
#
extra_pipeline_printout_graph_options = [
    "ignore_upstream_of_target",
    "skip_uptodate_tasks",
    "gnu_make_maximal_rebuild_mode",
    "test_all_task_for_update",
    "minimal_key_legend",
    "user_colour_scheme",
    "pipeline_name",
    "size",
    "dpi",
    "history_file",
    "checksum_level",
    "runtime_data",
]
extra_pipeline_printout_options = [
    "indent",
    "gnu_make_maximal_rebuild_mode",
    "checksum_level",
    "history_file",
    "wrap_width",
    "runtime_data"]


extra_pipeline_run_options = [
    "gnu_make_maximal_rebuild_mode",
    "runtime_data",
    "one_second_per_job",
    # exposed directly in command line
    #"touch_files_only"                  ,
    "history_file",
    "logger",
    "exceptions_terminate_immediately",
    "log_exceptions",
    "checksum_level",
    "multithread"]


# _________________________________________________________________________________________

#   get_extra_options_appropriate_for_command

# _________________________________________________________________________________________
def get_extra_options_appropriate_for_command(appropriate_option_names, extra_options):
    """
    Get extra options which are appropriate for
        pipeline_printout
        pipeline_printout_graph
        pipeline_run
    """

    appropriate_options = dict()
    for option_name in appropriate_option_names:
        if option_name in extra_options:
            appropriate_options[option_name] = extra_options[option_name]
    return appropriate_options


# _________________________________________________________________________________________

#   handle_verbose

# _________________________________________________________________________________________
def handle_verbose(options):
    """
    raw options.verbose is a list of specifiers
         '+'         : i.e. --verbose. This just increases the current verbosity value by 1
         '\d+'       : e.g. --verbose 6. This (re)sets the verbosity value
         '\d+:\d+'   : e.g. --verbose 7:-5 The second number is the verbose_abbreviated_path
        Set
            options.verbosity
            options.verbose_abbreviated_path

        Since the latter cannot be disabled via ignored_args (it never appears as a
            command line option), we do the next best thing
            by not overriding whatever the user sets

    """
    #
    #   verbosity specified manually or deliberately disabled: use that
    #
    if options.verbose is None or isinstance(options.verbose, int):
        #   verbose_abbreviated_path default to None unless set explicity by the user
        #       in which case we shall prudently not override it!
        #   verbose_abbreviated_path of None will be set to the default at
        #       pipeline_run() / pipeline_printout()
        if not hasattr(options, "verbose_abbreviated_path"):
            setattr(options, "verbose_abbreviated_path", None)
        return options
    #
    #   The user is having a laugh by passing in a string!
    #       wrap in list
    if isinstance(options.verbose, str):
        options.verbose = [options.verbose]
    #
    #
    curr_verbosity = 0
    curr_verbose_abbreviated_path = None
    import re
    match_regex = re.compile(r"(\+)|(\d+)(?::(\-?\d+))?")
    #
    #   Each verbosity specifier can be
    #       '+'         : i.e. --verbose. This just increases the current verbosity value by 1
    #       '\d+'       : e.g. --verbose 6. This (re)sets the verbosity value
    #       '\d+:\d+'   : e.g. --verbose 7:-5 The second number is the verbose_abbreviated_path
    #
    for vv in options.verbose:
        mm = match_regex.match(vv)
        if not mm:
            raise Exception(
                "error: verbosity argument is specified as --verbose INT or --verbose INT:INT. invalid value '%s'" % vv)
        if mm.group(1):
            curr_verbosity += 1
        else:
            curr_verbosity = int(mm.group(2))
            if mm.group(3):
                curr_verbose_abbreviated_path = int(mm.group(3))
    #
    # set verbose_abbreviated_path unless set explicity by the user
    #   in which case we shall prudently not override it!
    if not hasattr(options, "verbose_abbreviated_path"):
        setattr(options, "verbose_abbreviated_path",
                curr_verbose_abbreviated_path)
    options.verbose = curr_verbosity
    #
    return options


# _________________________________________________________________________________________

#   run

# _________________________________________________________________________________________
def run(options, **extra_options):
    """
    Take action depending on options
    extra_options are passed (as appropriate to the underlying functions
    Returns True if pipeline_run
    """

    #
    #   be very defensive: these options names are use below. Make sure they already
    #       exist in ``options`` , even if they have a value of None
    #
    for attr_name in ["just_print",
                      "verbose",
                      "flowchart",
                      "flowchart_format",
                      "target_tasks",
                      "forced_tasks",
                      "draw_horizontally",
                      "key_legend_in_graph",
                      "use_threads",
                      "jobs",
                      "recreate_database",
                      "touch_files_only",
                      "history_file"]:
        if not hasattr(options, attr_name):
            setattr(options, attr_name, None)

    #
    #   handle verbosity specification
    #
    #       the special attribute verbose_abbreviated_path is set in this function
    options = handle_verbose(options)

    #
    #   touch files or not
    #
    if options.recreate_database:
        touch_files_only = CHECKSUM_REGENERATE
    elif options.touch_files_only:
        touch_files_only = True
    else:
        touch_files_only = False

    if options.just_print:
        appropriate_options = get_extra_options_appropriate_for_command(
            extra_pipeline_printout_options, extra_options)
        task.pipeline_printout(sys.stdout, options.target_tasks, options.forced_tasks,
                               history_file=options.history_file,
                               verbose_abbreviated_path=options.verbose_abbreviated_path,
                               verbose=options.verbose, **appropriate_options)
        return False

    elif options.flowchart:
        appropriate_options = get_extra_options_appropriate_for_command(
            extra_pipeline_printout_graph_options, extra_options)
        task.pipeline_printout_graph(open(options.flowchart, "wb"),
                                     options.flowchart_format,
                                     options.target_tasks,
                                     options.forced_tasks,
                                     history_file=options.history_file,
                                     draw_vertically=not options.draw_horizontally,
                                     no_key_legend=not options.key_legend_in_graph,
                                     **appropriate_options)
        return False
    else:

        #
        #   turn on multithread if --use_threads specified and --jobs > 1
        #       ignore if manually specified
        #
        if (options.use_threads
                # ignore if manual override
                and not "multithread" in extra_options
                and options.jobs and options.jobs > 1):
            multithread = options.jobs
        elif "multithread" in extra_options:
            multithread = extra_options["multithread"]
            del extra_options["multithread"]
        else:
            multithread = None

        if not "logger" in extra_options:
            extra_options["logger"] = None
        if extra_options["logger"] == False:
            extra_options["logger"] = task.black_hole_logger
        elif extra_options["logger"] is None:
            extra_options["logger"] = task.stderr_logger
        appropriate_options = get_extra_options_appropriate_for_command(
            extra_pipeline_run_options, extra_options)
        task.pipeline_run(options.target_tasks,
                          options.forced_tasks,
                          multiprocess=options.jobs,
                          multithread=multithread,
                          verbose=options.verbose,
                          touch_files_only=touch_files_only,
                          history_file=options.history_file,
                          verbose_abbreviated_path=options.verbose_abbreviated_path,
                          **appropriate_options)
        return True



#!/usr/bin/env python
from __future__ import print_function
import unittest
from ruffus import pipeline_run, Pipeline, files

"""

    test_pausing.py

        test time.sleep keeping input files and output file times correct

"""

import os
import sys

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

try:
    import StringIO as io
except:
    import io as io

# use simplejson in place of json for python < 2.6
try:
    import json
except ImportError:
    import simplejson
    json = simplejson


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
a = []

#
#    task1
#


@files(a)
def task1():
    """
    First task
    """
    print("Task1", file=sys.stderr)


class t_save_to_str_logger:
    """
    Everything to stderr
    """

    def __init__(self):
        self.info_str = ""
        self.warning_str = ""
        self.debug_str = ""

    def info(self, message):
        self.info_str += message

    def warning(self, message):
        self.warning_str += message

    def debug(self, message):
        self.debug_str += message


class Test_task(unittest.TestCase):

    def test_task(self):

        save_to_str_logger = t_save_to_str_logger()
        pipeline_run(multiprocess=10,
                     logger=save_to_str_logger,
                     verbose=1,
                     pipeline="main")
        self.assertTrue("@files() was empty" in save_to_str_logger.warning_str)
        print("\n    Warning printed out correctly", file=sys.stderr)

    def test_newstyle_task(self):
        test_pipeline = Pipeline("test")
        test_pipeline.files(task1, a)

        save_to_str_logger = t_save_to_str_logger()
        test_pipeline.run(multiprocess=10,
                          logger=save_to_str_logger,
                          verbose=1)
        self.assertTrue("@files() was empty" in save_to_str_logger.warning_str)
        print("\n    Warning printed out correctly", file=sys.stderr)


#
#   Necessary to protect the "entry point" of the program under windows.
#       see: http://docs.python.org/library/multiprocessing.html#multiprocessing-programming
#
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import shutil
import unittest
from ruffus import pipeline_run, pipeline_printout, Pipeline, formatter, originate, subdivide, split
import time
import sys

"""

    test_split_subdivide_checkpointing.py


"""
import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Each time the pipeline is FORCED to rerun,
#       More files are created for each task


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


@originate([tempdir + 'start'])
def make_start(outfile):
    """
    -> start
    """
    open(outfile, 'w').close()


@split(make_start, tempdir + '*.split')
def split_start(infiles, outfiles):
    """
    -> XXX.split
        where XXX = 0 .. N,
              N   = previous N + 1
    """

    # split always runs exactly one job (unlike @subdivide)
    # So it implicitly combines all its inputs before running and generating multiple output
    # @originate generates multiple output so the input for @split is a list...
    infile = infiles[0]

    # clean up previous
    for f in outfiles:
        os.unlink(f)

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    #
    #   Create more files than the previous invocation
    #
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    n_to_produce = len(outfiles) + 1
    for i in range(n_to_produce):
        f = '{}{}.split'.format(tempdir, i)
        open(f, 'a').close()


@subdivide(split_start, formatter(), tempdir + '{basename[0]}_*.subdivided', tempdir + '{basename[0]}')
def subdivide_start(infile, outfiles, infile_basename):
    # cleanup existing
    for f in outfiles:
        os.unlink(f)

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    #
    #   Create more files than the previous invocation
    #
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    n_to_produce = len(outfiles) + 1
    output_file_names = []
    for i in range(n_to_produce):
        output_file_name = '{}_{}.subdivided'.format(infile_basename, i)
        open(output_file_name, 'a').close()
        output_file_names.append(output_file_name)


TEST_VERBOSITY = 0


class Test_ruffus(unittest.TestCase):

    def tearDown(self):
        # only tear down if not throw exception so we can debug?
        try:
            shutil.rmtree(tempdir)
        except:
            pass

    def setUp(self):
        try:
            shutil.rmtree(tempdir)
        except:
            pass
        os.makedirs(tempdir)

    def check_file_exists_or_not_as_expected(self, expected_files, not_expected_files):
        """
        Check if files exist / not exist
        """
        for ee in expected_files:
            if not os.path.exists(tempdir + ee):
                raise Exception("Expected file %s" % (tempdir + ee))
        for ne in not_expected_files:
            if os.path.exists(tempdir + ne):
                raise Exception("Unexpected file %s" % (tempdir + ne))

    def test_newstyle_ruffus(self):

        test_pipeline = Pipeline("test")
        test_pipeline.originate(task_func=make_start,
                                output=[tempdir + 'start'])
        test_pipeline.split(task_func=split_start,
                            input=make_start, output=tempdir + '*.split')
        test_pipeline.subdivide(task_func=subdivide_start, input=split_start, filter=formatter(
        ), output=tempdir + '{basename[0]}_*.subdivided', extras=[tempdir + '{basename[0]}'])

        expected_files_after_1_runs = ["start", "0.split", "0_0.subdivided"]
        expected_files_after_2_runs = [
            "1.split", "0_1.subdivided", "1_0.subdivided"]
        expected_files_after_3_runs = [
            "2.split", "0_2.subdivided", "1_1.subdivided", "2_0.subdivided"]
        expected_files_after_4_runs = [
            "3.split", "0_3.subdivided", "1_2.subdivided", "2_1.subdivided", "3_0.subdivided"]

        print("     1 Run pipeline normally...")
        test_pipeline.run(multiprocess=10, verbose=TEST_VERBOSITY)
        self.check_file_exists_or_not_as_expected(expected_files_after_1_runs,
                                                  expected_files_after_2_runs)
        print("     2 Check that running again does nothing. (All up to date).")
        test_pipeline.run(multiprocess=10, verbose=TEST_VERBOSITY)
        self.check_file_exists_or_not_as_expected(expected_files_after_1_runs,
                                                  expected_files_after_2_runs)
        time.sleep(2)

        print("     3 Running again with forced tasks to generate more files...")
        test_pipeline.run(forcedtorun_tasks=[
                          "test::make_start"], multiprocess=10, verbose=TEST_VERBOSITY)
        self.check_file_exists_or_not_as_expected(expected_files_after_1_runs
                                                  + expected_files_after_2_runs,
                                                  expected_files_after_3_runs)
        print("     4 Check that running again does nothing. (All up to date).")
        test_pipeline.run(multiprocess=10, verbose=TEST_VERBOSITY)
        self.check_file_exists_or_not_as_expected(expected_files_after_1_runs
                                                  + expected_files_after_2_runs,
                                                  expected_files_after_3_runs)
        time.sleep(2)

        print("     5 Running again with forced tasks to generate even more files...")
        test_pipeline.run(forcedtorun_tasks=make_start,
                          multiprocess=10, verbose=TEST_VERBOSITY)
        self.check_file_exists_or_not_as_expected(expected_files_after_1_runs
                                                  + expected_files_after_2_runs
                                                  + expected_files_after_3_runs,
                                                  expected_files_after_4_runs)
        print("     6 Check that running again does nothing. (All up to date).")
        test_pipeline.run(multiprocess=10, verbose=TEST_VERBOSITY)
        self.check_file_exists_or_not_as_expected(expected_files_after_1_runs
                                                  + expected_files_after_2_runs
                                                  + expected_files_after_3_runs,
                                                  expected_files_after_4_runs)

    def test_ruffus(self):

        expected_files_after_1_runs = ["start", "0.split", "0_0.subdivided"]
        expected_files_after_2_runs = [
            "1.split", "0_1.subdivided", "1_0.subdivided"]
        expected_files_after_3_runs = [
            "2.split", "0_2.subdivided", "1_1.subdivided", "2_0.subdivided"]
        expected_files_after_4_runs = [
            "3.split", "0_3.subdivided", "1_2.subdivided", "2_1.subdivided", "3_0.subdivided"]

        print("     1 Run pipeline normally...")
        pipeline_run(multiprocess=10, verbose=TEST_VERBOSITY, pipeline="main")
        self.check_file_exists_or_not_as_expected(expected_files_after_1_runs,
                                                  expected_files_after_2_runs)
        print("     2 Check that running again does nothing. (All up to date).")
        pipeline_run(multiprocess=10, verbose=TEST_VERBOSITY, pipeline="main")
        self.check_file_exists_or_not_as_expected(expected_files_after_1_runs,
                                                  expected_files_after_2_runs)

        time.sleep(2)

        print("     3 Running again with forced tasks to generate more files...")
        pipeline_run(forcedtorun_tasks=[
                     make_start], multiprocess=10, verbose=TEST_VERBOSITY, pipeline="main")
        self.check_file_exists_or_not_as_expected(expected_files_after_1_runs
                                                  + expected_files_after_2_runs,
                                                  expected_files_after_3_runs)
        print("     4 Check that running again does nothing. (All up to date).")
        pipeline_run(multiprocess=10, verbose=TEST_VERBOSITY, pipeline="main")
        self.check_file_exists_or_not_as_expected(expected_files_after_1_runs
                                                  + expected_files_after_2_runs,
                                                  expected_files_after_3_runs)
        time.sleep(2)

        print("     5 Running again with forced tasks to generate even more files...")
        pipeline_run(forcedtorun_tasks=[
                     make_start], multiprocess=10, verbose=TEST_VERBOSITY, pipeline="main")
        self.check_file_exists_or_not_as_expected(expected_files_after_1_runs
                                                  + expected_files_after_2_runs
                                                  + expected_files_after_3_runs,
                                                  expected_files_after_4_runs)
        time.sleep(2)
        print("     6 Check that running again does nothing. (All up to date).")
        pipeline_run(multiprocess=10, verbose=TEST_VERBOSITY, pipeline="main")
        self.check_file_exists_or_not_as_expected(expected_files_after_1_runs
                                                  + expected_files_after_2_runs
                                                  + expected_files_after_3_runs,
                                                  expected_files_after_4_runs)


if __name__ == '__main__':
    if len(sys.argv) > 1 and "-v" == sys.argv.pop():
        TEST_VERBOSITY = 7
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import unittest
from ruffus import task
from ruffus import ruffus_exceptions

"""
    test_task_misc.py
"""

import os
import sys

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


class Test_needs_update_check_directory_missing(unittest.TestCase):

    def setUp(self):
        """
        Create temp directory and temp file
        """
        import tempfile

        # test_file =tempfile.NamedTemporaryFile(delete=False)
        # self.tempfile = test_file.name
        # test_file.close()
        fh, self.tempfile = tempfile.mkstemp(suffix='.dot')
        os.fdopen(fh, "w").close()
        self.directory = tempfile.mkdtemp(prefix='testing_tmp')

    def tearDown(self):
        """
        delete files
        """
        os.unlink(self.tempfile)
        os.removedirs(self.directory)

    def test_up_to_date(self):
        #
        #   lists of files
        #

        self.assertTrue(
            not task.needs_update_check_directory_missing([self.directory])[0])
        self.assertTrue(task.needs_update_check_directory_missing(
            ["missing directory"])[0])
        self.assertRaises(ruffus_exceptions.error_not_a_directory,
                          task.needs_update_check_directory_missing, [self.tempfile])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""

    exceptions.py

"""

################################################################################
#
#   exceptions.py
#
#
#   Copyright (c) 10/9/2009 Leo Goodstadt
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#################################################################################

import sys
import os
from collections import defaultdict


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Exceptions


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

# if __name__ != '__main__':
#    import task

class error_task(Exception):
    def __init__(self, *errmsg):
        Exception.__init__(self, *errmsg)

        # list of associated tasks
        self.tasks = set()

        # error message
        self.main_msg = ""

    def get_main_msg(self):
        """
        Make main message with lists of task names
        Prefix with new lines for added emphasis
        """
        # turn tasks names into 'def xxx(...): format
        task_names = "\n".join("task = %r" % t._name for t in self.tasks)
        if len(self.main_msg):
            return "\n\n" + self.main_msg + " for\n\n%s\n" % task_names
        else:
            return "\n\n%s\n" % task_names

    def __str__(self):
        # indent
        msg = self.get_main_msg() + " ".join(map(str, self.args))
        return "    " + msg.replace("\n", "\n    ")

    def specify_task(self, task, main_msg):
        self.tasks.add(task)
        self.main_msg = main_msg
        return self


class error_task_contruction(error_task):
    """
    Exceptions when contructing pipeline tasks
    """

    def __init__(self, task, main_msg, *errmsg):
        error_task.__init__(self, *errmsg)
        self.specify_task(task, main_msg)


class RethrownJobError(error_task):
    """
    Wrap up one or more exceptions rethrown across process boundaries

        See multiprocessor.Server.handle_request/serve_client for an analogous function
    """

    def __init__(self, job_exceptions=[]):
        error_task.__init__(self)
        self.job_exceptions = list(job_exceptions)

    def __len__(self):
        return len(self.job_exceptions)

    def append(self, job_exception):
        self.job_exceptions.append(job_exception)

    def task_to_func_name(self, task_name):
        if "mkdir " in task_name:
            return task_name

        return "def %s(...):" % task_name.replace("__main__.", "")

    def get_nth_exception_str(self, nn=-1):
        if nn == -1:
            nn = len(self.job_exceptions) - 1
        task_name, job_name, exception_name, exception_value, exception_stack = self.job_exceptions[
            nn]
        message = "\nException #%d\n" % (nn + 1)
        message += "  '%s%s' raised in ...\n" % (exception_name, exception_value)
        if task_name:
            message += "   Task = %s\n   %s\n\n" % (self.task_to_func_name(task_name),
                                                    job_name)
        message += "%s\n" % (exception_stack, )
        return message.replace("\n", "\n    ")

    def __str__(self):
        message = ["\nOriginal exception%s:\n" %
                   ("s" if len(self.job_exceptions) > 1 else "")]
        for ii in range(len(self.job_exceptions)):
            message += self.get_nth_exception_str(ii)
        #
        #   For each exception:
        #       turn original exception stack message into an indented string
        #
        return (self.get_main_msg()).replace("\n", "\n    ") + "".join(message)


class error_input_file_does_not_match(error_task):
    pass


class fatal_error_input_file_does_not_match(error_task):
    pass


class task_FilesArgumentsError(error_task):
    pass


class task_FilesreArgumentsError(error_task):
    pass


class MissingInputFileError(error_task):
    pass


class JobSignalledBreak(error_task):
    pass


class JobSignalledSuspend(error_task):
    pass


class JobSignalledResume(error_task):
    pass


class JobFailed(error_task):
    pass


class PostTaskArgumentError(error_task):
    pass


class JobsLimitArgumentError(error_task):
    pass


class error_task_get_output(error_task_contruction):
    pass


class error_task_transform_inputs_multiple_args(error_task_contruction):
    pass


class error_task_transform(error_task_contruction):
    pass


class error_task_product(error_task_contruction):
    pass


class error_task_mkdir(error_task_contruction):
    pass


class error_task_permutations(error_task_contruction):
    pass


class error_task_combinations(error_task_contruction):
    pass


class error_task_combinations_with_replacement(error_task_contruction):
    pass


class error_task_merge(error_task_contruction):
    pass


class error_task_subdivide(error_task_contruction):
    pass


class error_task_originate(error_task_contruction):
    pass


class error_task_collate(error_task_contruction):
    pass


class error_task_collate_inputs_multiple_args(error_task_contruction):
    pass


class error_task_split(error_task_contruction):
    pass


class error_task_files_re(error_task_contruction):
    pass


class error_task_files(error_task_contruction):
    pass


class error_task_parallel(error_task_contruction):
    pass


class error_making_directory(error_task):
    pass


class error_duplicate_task_name(error_task):
    pass


class error_decorator_args(error_task):
    pass


class error_task_name_lookup_failed(error_task):
    pass


class error_task_decorator_takes_no_args(error_task):
    pass


class error_function_is_not_a_task(error_task):
    pass


class error_ambiguous_task(error_task):
    pass


class error_not_a_pipeline(error_task):
    pass


class error_circular_dependencies(error_task):
    pass


class error_not_a_directory(error_task):
    pass


class error_missing_output(error_task):
    pass


class error_job_signalled_interrupt(error_task):
    pass


class error_node_not_task(error_task):
    pass


class error_missing_runtime_parameter(error_task):
    pass


class error_unescaped_regular_expression_forms(error_task):
    pass


class error_checksum_level(error_task):
    pass


class error_missing_args(error_task):
    pass


class error_too_many_args(error_task):
    pass


class error_inputs_multiple_args(error_task):
    pass


class error_set_input(error_task):
    pass


class error_set_output(error_task):
    pass


class error_no_head_tasks(error_task):
    pass


class error_no_tail_tasks(error_task):
    pass


class error_executable_str(error_task):
    pass


class error_extras_wrong_type(error_task):
    pass


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Testing
# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
if __name__ == '__main__':
    import unittest

    #
    #   minimal task object to test exceptions
    #
    class task:
        class Task (object):
            """
            dummy task
            """
            _action_mkdir = 1

            def __init__(self, _name,  _action_type=0):
                self._action_type = _action_type
                self._name = _name

    class Test_exceptions(unittest.TestCase):

        #       self.assertEqual(self.seq, range(10))
        #       self.assert_(element in self.seq)
        #       self.assertRaises(ValueError, random.sample, self.seq, 20)

        def test_error_task(self):
            """
                test
            """
            fake_task1 = task.Task("task1")
            fake_task2 = task.Task("task2")
            fake_mkdir_task3 = task.Task("task3", task.Task._action_mkdir)
            fake_mkdir_task4 = task.Task("task4", task.Task._action_mkdir)
            e = error_task()
            e.specify_task(fake_task1, "Some message 0")
            e.specify_task(fake_task2, "Some message 1")
            e.specify_task(fake_mkdir_task3, "Some message 2")
            e.specify_task(fake_mkdir_task4, "Some message 3")
            self.assertEqual(str(e),
                             """

    Some message 3 for

    'def task1(...):'
    'def task2(...):'
    task3
    task4
    """)

        def test_RethrownJobError(self):
            """
                test
            """
            #job_name, exception_name, exception_value, exception_stack
            exception_data = [
                [
                    "task1",
                    "[[temp_branching_dir/a.2, a.1] -> temp_branching_dir/a.3]",
                    "ruffus.task.MissingInputFileError",
                    "(instance value)",
                    "Traceback (most recent call last):\n  File \"what.file.py\", line 333, in some_func\n  somecode(sfasf)\n"
                ],
                [
                    "task1",
                    "[None -> [temp_branching_dir/a.1, temp_branching_dir/b.1, temp_branching_dir/c.1]]",
                    "exceptions.ZeroDivisionError:",
                    "(1)",
                    "Traceback (most recent call last):\n  File \"anotherfile.py\", line 345, in other_func\n  badcode(rotten)\n"
                ]

            ]
            e = RethrownJobError(exception_data)
            fake_task1 = task.Task("task1")
            fake_task2 = task.Task("task2")
            fake_mkdir_task3 = task.Task("task3", task.Task._action_mkdir)
            fake_mkdir_task4 = task.Task("task4", task.Task._action_mkdir)
            e.specify_task(fake_task1, "Exceptions running jobs")
            e.specify_task(fake_task2, "Exceptions running jobs")
            e.specify_task(fake_mkdir_task3, "Exceptions running jobs")
            e.specify_task(fake_mkdir_task4, "Exceptions running jobs")
            self.assertEqual(str(e),
                             """

    Exceptions running jobs for

    'def task1(...):'
    'def task2(...):'
    task3
    task4

    Original exceptions:

    Exception #1
    ruffus.task.MissingInputFileError(instance value):
    for task1.[[temp_branching_dir/a.2, a.1] -> temp_branching_dir/a.3]

    Traceback (most recent call last):
      File "what.file.py", line 333, in some_func
      somecode(sfasf)


    Exception #2
    exceptions.ZeroDivisionError:(1):
    for task1.[None -> [temp_branching_dir/a.1, temp_branching_dir/b.1, temp_branching_dir/c.1]]

    Traceback (most recent call last):
      File "anotherfile.py", line 345, in other_func
      badcode(rotten)

    """)


#
#   debug code not run if called as a module
#
if __name__ == '__main__':
    if sys.argv.count("--debug"):
        sys.argv.remove("--debug")
    unittest.main()



################################################################################
#
#   __init__.py
#
#
#   Copyright (c) 10/9/2009 Leo Goodstadt
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#################################################################################
# pipeline functions
from .task import Pipeline, Task

# pipeline functions
from .task import pipeline_printout, pipeline_printout_graph, pipeline_run, pipeline_get_task_names, register_cleanup, \
	lookup_pipeline

# decorators
from .task import originate, split, subdivide, transform, merge, collate, follows

# esoteric decorators
from .task import files, parallel

# filter / indicators
from .task import touch_file
from .file_name_parameters import suffix, regex, formatter, inputs, add_inputs

# deprecated
from .task import files_re
from .ruffus_utility import combine

from .task import check_if_uptodate, active_if, jobs_limit, graphviz, mkdir, posttask
from .ruffus_utility import output_from, runtime_parameter
from .task import stderr_logger, black_hole_logger
from .ruffus_exceptions import JobSignalledBreak
from .graph import graph_colour_demo_printout
from .file_name_parameters import needs_update_check_modify_time
from . import cmdline
from . import combinatorics

# output_dependency_tree_in_dot_format, output_dependency_tree_key_in_dot_format
from . import ruffus_version
__version__ = ruffus_version.__version



#!/usr/bin/env python
from __future__ import print_function
from ruffus.ruffus_utility import *
from ruffus import *

################################################################################
#
#   test_ruffus_utility_parse_task_arguments.py
#
#################################################################################
"""
    test_ruffus_utility.py
"""

import unittest

import os
import sys

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# _________________________________________________________________________________________

#   Test_parse_transform_args

# _________________________________________________________________________________________

class Test_parse_transform_args (unittest.TestCase):

    def test_parse_transform_args(self):
        expected_arguments = ["input", "filter",
                              "modify_inputs", "output", "extras"]

        empty_unnamed_arguments = []
        empty_named_arguments = {}
        orig_unnamed_arguments = [
            "*.txt", suffix(".txt"), ".result", 1, 2, 3, 4]
        task_description = "@transform(%s)\ndef myfunc(...)\n"
        expected_results = {'input': orig_unnamed_arguments[0],
                            'filter': orig_unnamed_arguments[1],
                            'output': orig_unnamed_arguments[2],
                            'extras': orig_unnamed_arguments[3:],
                            'named_extras': {},
                            'modify_inputs_mode': 2,
                            'modify_inputs': None}
        add_inputs_expected_results = {'input': orig_unnamed_arguments[0],
                                       'filter': orig_unnamed_arguments[1],
                                       'output': orig_unnamed_arguments[2],
                                       'extras': orig_unnamed_arguments[3:],
                                       'named_extras': {},
                                       'modify_inputs_mode': 0,
                                       'modify_inputs': ("a.test", "b.test")}
        replace_inputs_expected_results = {'input': orig_unnamed_arguments[0],
                                           'filter': orig_unnamed_arguments[1],
                                           'output': orig_unnamed_arguments[2],
                                           'extras': orig_unnamed_arguments[3:],
                                           'named_extras': {},
                                           'modify_inputs_mode': 1,
                                           'modify_inputs': ("a.test", "b.test")}

        # Error: empty list
        with self.assertRaises(error_missing_args):
            parse_task_arguments(
                empty_named_arguments, empty_named_arguments, expected_arguments, task_description)

        # parse complete correctly
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {}, expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # Error: missing argument
        unnamed_arguments = orig_unnamed_arguments[0:1]
        with self.assertRaises(error_missing_args):
            results = parse_task_arguments(unnamed_arguments,
                                           {}, expected_arguments, task_description)

        # parse almost complete and rescued with named parameter
        unnamed_arguments = orig_unnamed_arguments[0:2]
        results = parse_task_arguments(unnamed_arguments,
                                       {'output': orig_unnamed_arguments[2],
                                        'extras': orig_unnamed_arguments[3:]},
                                       expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # All named parameters
        results = parse_task_arguments([],
                                       {'input': orig_unnamed_arguments[0],
                                        'filter': orig_unnamed_arguments[1],
                                        'output': orig_unnamed_arguments[2],
                                        'extras': orig_unnamed_arguments[3:],
                                        },
                                       expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # filter not regex suffix or formatter
        with self.assertRaises(TypeError):
            results = parse_task_arguments([],
                                           {'input': orig_unnamed_arguments[0],
                                            'filter': "a",
                                            'output': orig_unnamed_arguments[2],
                                            'extras': orig_unnamed_arguments[3:],
                                            },
                                           expected_arguments, task_description)

        # Error: Unknown named arguments
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments([],
                                           {'input': orig_unnamed_arguments[0],
                                            'filter': orig_unnamed_arguments[1],
                                            'output': orig_unnamed_arguments[2],
                                            'what': orig_unnamed_arguments[3:]
                                            },
                                           expected_arguments, task_description)

        # Error: Duplicate named arguments
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {'input': orig_unnamed_arguments[0],
                                            'extras': orig_unnamed_arguments[3:],
                                            'named_extras': {},
                                            },
                                           expected_arguments, task_description)

        # add_inputs correct via named
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {"add_inputs": ("a.test", "b.test")}, expected_arguments, task_description)
        self.assertEqual(results, add_inputs_expected_results)

        # add_inputs correct via named and paranoid add_inputs wrapping
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {"add_inputs": add_inputs("a.test", "b.test")}, expected_arguments, task_description)
        self.assertEqual(results, add_inputs_expected_results)

        # add_inputs correct via unnamed
        unnamed_arguments = list(orig_unnamed_arguments)
        unnamed_arguments.insert(2, add_inputs("a.test", "b.test"))
        results = parse_task_arguments(unnamed_arguments, {},
                                       expected_arguments, task_description)
        self.assertEqual(results, add_inputs_expected_results)

        # replace_inputs correct via named
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {"replace_inputs": ("a.test", "b.test")}, expected_arguments, task_description)
        self.assertEqual(results, replace_inputs_expected_results)

        # replace_inputs correct via named and paranoid inputs() wrapping
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {"replace_inputs": inputs(("a.test", "b.test"))}, expected_arguments, task_description)
        self.assertEqual(results, replace_inputs_expected_results)

        # replace_inputs correct via unnamed
        unnamed_arguments = list(orig_unnamed_arguments)
        unnamed_arguments.insert(2, inputs(("a.test", "b.test")))
        results = parse_task_arguments(unnamed_arguments, {},
                                       expected_arguments, task_description)
        self.assertEqual(results, replace_inputs_expected_results)

        # Error: both add_inputs and replace_inputs via named
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {"replace_inputs": ("a.test", "b.test"),
                                            "add_inputs": ("a.test", "b.test")}, expected_arguments, task_description)

        # Error: both add_inputs and replace_inputs via named / unnamed
        with self.assertRaises(error_too_many_args):
            unnamed_arguments = list(orig_unnamed_arguments)
            unnamed_arguments.insert(2, inputs(("a.test", "b.test")))
            results = parse_task_arguments(unnamed_arguments, {"add_inputs": ("a.test", "b.test")},
                                           expected_arguments, task_description)

        # Error: both add_inputs and replace_inputs via named / unnamed
        with self.assertRaises(error_too_many_args):
            unnamed_arguments = list(orig_unnamed_arguments)
            unnamed_arguments.insert(2, add_inputs("a.test", "b.test"))
            results = parse_task_arguments(unnamed_arguments, {"replace_inputs": ("a.test", "b.test")},
                                           expected_arguments, task_description)

        # Error: wrong number of arguments
        with self.assertRaises(error_inputs_multiple_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {"replace_inputs": inputs("a.test", "b.test")}, expected_arguments, task_description)

        with self.assertRaises(error_inputs_multiple_args):
            unnamed_arguments = orig_unnamed_arguments[0:2] + [
                inputs("a.test", "b.test")] + orig_unnamed_arguments[2:]
            results = parse_task_arguments(
                unnamed_arguments, {}, expected_arguments, task_description)

        # Error: no arguments
        with self.assertRaises(error_inputs_multiple_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {"replace_inputs": inputs()}, expected_arguments, task_description)


# ________________________________________________________________________________________________________

#   Test_parse_product_args

# ________________________________________________________________________________________________________
class Test_parse_product_args (unittest.TestCase):
    """
    Make sure (input, filter, input2, filter2, input3, filter3,..., output, extras...) works
    for @product
    """

    def test_parse_product_args(self):
        self.maxDiff = None
        expected_arguments = ["input", "filter",
                              "inputN", "modify_inputs", "output", "extras"]

        empty_unnamed_arguments = []
        empty_named_arguments = {}
        orig_unnamed_arguments = ["*.txt", formatter(".txt"), "*.contig", formatter(
        ), "*.genome", formatter(), "{basename[0][0]}_{basename[1][0]}.result", 1, 2, 3, 4]
        task_description = "@product(%s)\ndef myfunc(...)\n"
        expected_results = {'input': [orig_unnamed_arguments[0], orig_unnamed_arguments[2], orig_unnamed_arguments[4]],
                            'filter': [orig_unnamed_arguments[1], orig_unnamed_arguments[3], orig_unnamed_arguments[5]],
                            'output': orig_unnamed_arguments[6],
                            'extras': orig_unnamed_arguments[7:],
                            'named_extras': {},
                            'modify_inputs_mode': 2,
                            'modify_inputs': None}
        add_inputs_expected_results = {'input': [orig_unnamed_arguments[0], orig_unnamed_arguments[2], orig_unnamed_arguments[4]],
                                       'filter': [orig_unnamed_arguments[1], orig_unnamed_arguments[3], orig_unnamed_arguments[5]],
                                       'output': orig_unnamed_arguments[6],
                                       'extras': orig_unnamed_arguments[7:],
                                       'named_extras': {},
                                       'modify_inputs_mode': 0,
                                       'modify_inputs': ("a.test", "b.test")}
        replace_inputs_expected_results = {'input': [orig_unnamed_arguments[0], orig_unnamed_arguments[2], orig_unnamed_arguments[4]],
                                           'filter': [orig_unnamed_arguments[1], orig_unnamed_arguments[3], orig_unnamed_arguments[5]],
                                           'output': orig_unnamed_arguments[6],
                                           'extras': orig_unnamed_arguments[7:],
                                           'named_extras': {},
                                           'modify_inputs_mode': 1,
                                           'modify_inputs': ("a.test", "b.test")}

        # Error: empty list
        with self.assertRaises(error_missing_args):
            parse_task_arguments(
                empty_named_arguments, empty_named_arguments, expected_arguments, task_description)

        # parse complete correctly
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {}, expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # Error: missing argument
        unnamed_arguments = orig_unnamed_arguments[0:6]
        with self.assertRaises(error_missing_args):
            results = parse_task_arguments(unnamed_arguments,
                                           {}, expected_arguments, task_description)

        # parse almost complete and rescued with named parameter
        unnamed_arguments = orig_unnamed_arguments[0:6]
        results = parse_task_arguments(unnamed_arguments,
                                       {'output': expected_results['output'],
                                        'extras': expected_results['extras']},
                                       expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # All named parameters
        results = parse_task_arguments([],
                                       {'input': orig_unnamed_arguments[0],
                                        'filter': orig_unnamed_arguments[1],
                                        'input2': orig_unnamed_arguments[2],
                                        'filter2': orig_unnamed_arguments[3],
                                        'input3': orig_unnamed_arguments[4],
                                        'filter3': orig_unnamed_arguments[5],
                                        'output': orig_unnamed_arguments[6],
                                        'extras': orig_unnamed_arguments[7:],
                                        },
                                       expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # Error: Unknown named arguments
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments([],
                                           {'input': orig_unnamed_arguments[0],
                                            'filter': orig_unnamed_arguments[1],
                                            'input2': orig_unnamed_arguments[2],
                                            'filter2': orig_unnamed_arguments[3],
                                            'input3': orig_unnamed_arguments[4],
                                            'filter3': orig_unnamed_arguments[5],
                                            'output': orig_unnamed_arguments[6],
                                            'what': orig_unnamed_arguments[7:]
                                            },
                                           expected_arguments, task_description)

        # Error: Duplicate named arguments
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {'input': orig_unnamed_arguments[0],
                                            'extras': orig_unnamed_arguments[7:],
                                            },
                                           expected_arguments, task_description)

        # add_inputs correct via named
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {"add_inputs": ("a.test", "b.test")}, expected_arguments, task_description)
        self.assertEqual(results, add_inputs_expected_results)

        # add_inputs correct via named and paranoid add_inputs wrapping
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {"add_inputs": add_inputs("a.test", "b.test")}, expected_arguments, task_description)
        self.assertEqual(results, add_inputs_expected_results)

        # add_inputs correct via unnamed
        unnamed_arguments = list(orig_unnamed_arguments)
        unnamed_arguments.insert(6, add_inputs("a.test", "b.test"))
        results = parse_task_arguments(unnamed_arguments, {},
                                       expected_arguments, task_description)
        self.assertEqual(results, add_inputs_expected_results)

        # replace_inputs correct via named
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {"replace_inputs": ("a.test", "b.test")}, expected_arguments, task_description)
        self.assertEqual(results, replace_inputs_expected_results)

        # replace_inputs correct via named and paranoid inputs() wrapping
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {"replace_inputs": inputs(("a.test", "b.test"))}, expected_arguments, task_description)
        self.assertEqual(results, replace_inputs_expected_results)

        # replace_inputs correct via unnamed
        unnamed_arguments = list(orig_unnamed_arguments)
        unnamed_arguments.insert(6, inputs(("a.test", "b.test")))
        results = parse_task_arguments(unnamed_arguments, {},
                                       expected_arguments, task_description)
        self.assertEqual(results, replace_inputs_expected_results)

        # Error: both add_inputs and replace_inputs via named
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {"replace_inputs": ("a.test", "b.test"),
                                            "add_inputs": ("a.test", "b.test")}, expected_arguments, task_description)

        # Error: both add_inputs and replace_inputs via named / unnamed
        with self.assertRaises(error_too_many_args):
            unnamed_arguments = list(orig_unnamed_arguments)
            unnamed_arguments.insert(6, inputs(("a.test", "b.test")))
            results = parse_task_arguments(unnamed_arguments, {"add_inputs": ("a.test", "b.test")},
                                           expected_arguments, task_description)

        # Error: both add_inputs and replace_inputs via named / unnamed
        with self.assertRaises(error_too_many_args):
            unnamed_arguments = list(orig_unnamed_arguments)
            unnamed_arguments.insert(6, add_inputs("a.test", "b.test"))
            results = parse_task_arguments(unnamed_arguments, {"replace_inputs": ("a.test", "b.test")},
                                           expected_arguments, task_description)

        # Error: wrong number of arguments
        with self.assertRaises(error_inputs_multiple_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {"replace_inputs": inputs("a.test", "b.test")}, expected_arguments, task_description)

        # Error: no arguments
        with self.assertRaises(error_inputs_multiple_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {"replace_inputs": inputs()}, expected_arguments, task_description)


# ________________________________________________________________________________________________________

#   Test_parse_combinatorics_args

# ________________________________________________________________________________________________________
class Test_parse_combinatorics_args (unittest.TestCase):
    """
    Make sure (input, filter, tuple_size, output, extras...) works
    for @combinations
    """

    def test_parse_combinations_args(self):
        expected_arguments = ["input", "filter",
                              "tuple_size", "modify_inputs", "output", "extras"]

        empty_unnamed_arguments = []
        empty_named_arguments = {}
        orig_unnamed_arguments = [
            "*.txt", suffix(".txt"), 5, ".result", 1, 2, 3, 4]
        task_description = "@combinations(%s)\ndef myfunc(...)\n"
        expected_results = {'input': orig_unnamed_arguments[0],
                            'filter': orig_unnamed_arguments[1],
                            'tuple_size': orig_unnamed_arguments[2],
                            'output': orig_unnamed_arguments[3],
                            'extras': orig_unnamed_arguments[4:],
                            'named_extras': {},
                            'modify_inputs_mode': 2,
                            'modify_inputs': None}
        add_inputs_expected_results = {'input': orig_unnamed_arguments[0],
                                       'filter': orig_unnamed_arguments[1],
                                       'tuple_size': orig_unnamed_arguments[2],
                                       'output': orig_unnamed_arguments[3],
                                       'extras': orig_unnamed_arguments[4:],
                                       'named_extras': {},
                                       'modify_inputs_mode': 0,
                                       'modify_inputs': ("a.test", "b.test")}
        replace_inputs_expected_results = {'input': orig_unnamed_arguments[0],
                                           'filter': orig_unnamed_arguments[1],
                                           'tuple_size': orig_unnamed_arguments[2],
                                           'output': orig_unnamed_arguments[3],
                                           'extras': orig_unnamed_arguments[4:],
                                           'named_extras': {},
                                           'modify_inputs_mode': 1,
                                           'modify_inputs': ("a.test", "b.test")}

        # Error: empty list
        with self.assertRaises(error_missing_args):
            parse_task_arguments(
                empty_named_arguments, empty_named_arguments, expected_arguments, task_description)

        # parse complete correctly
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {}, expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # Error tuple_size not int
        unnamed_arguments = orig_unnamed_arguments[:]
        unnamed_arguments[2] = 'a'
        with self.assertRaises(TypeError):
            results = parse_task_arguments(unnamed_arguments,
                                           {}, expected_arguments, task_description)

        # Error: missing argument
        unnamed_arguments = orig_unnamed_arguments[0:2]
        with self.assertRaises(error_missing_args):
            results = parse_task_arguments(unnamed_arguments,
                                           {}, expected_arguments, task_description)

        # parse almost complete and rescued with named parameter
        unnamed_arguments = orig_unnamed_arguments[0:2]
        results = parse_task_arguments(unnamed_arguments,
                                       {
                                           'tuple_size': orig_unnamed_arguments[2],
                                           'output': orig_unnamed_arguments[3],
                                           'extras': orig_unnamed_arguments[4:]},
                                       expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # All named parameters
        unnamed_arguments = orig_unnamed_arguments[0:2]
        results = parse_task_arguments([],
                                       {'input': orig_unnamed_arguments[0],
                                        'filter': orig_unnamed_arguments[1],
                                        'tuple_size': orig_unnamed_arguments[2],
                                        'output': orig_unnamed_arguments[3],
                                        'extras': orig_unnamed_arguments[4:],
                                        },
                                       expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # Error tuple_size not int
        unnamed_arguments = orig_unnamed_arguments[0:2]
        with self.assertRaises(TypeError):
            results = parse_task_arguments([],
                                           {'input': orig_unnamed_arguments[0],
                                            'filter': orig_unnamed_arguments[1],
                                            'tuple_size': "a",
                                            'output': orig_unnamed_arguments[3],
                                            'extras': orig_unnamed_arguments[4:],
                                            },
                                           expected_arguments, task_description)

        # Error: Unknown named arguments
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments([],
                                           {'input': orig_unnamed_arguments[0],
                                            'filter': orig_unnamed_arguments[1],
                                            'tuple_size': orig_unnamed_arguments[2],
                                            'output': orig_unnamed_arguments[3],
                                            'what': orig_unnamed_arguments[4:]
                                            },
                                           expected_arguments, task_description)

        # Error: Duplicate named arguments
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {'input': orig_unnamed_arguments[0],
                                            'extras': orig_unnamed_arguments[3:],
                                            },
                                           expected_arguments, task_description)

        # add_inputs correct via named
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {"add_inputs": ("a.test", "b.test")}, expected_arguments, task_description)
        self.assertEqual(results, add_inputs_expected_results)

        # add_inputs correct via named and paranoid add_inputs wrapping
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {"add_inputs": add_inputs("a.test", "b.test")}, expected_arguments, task_description)
        self.assertEqual(results, add_inputs_expected_results)

        # add_inputs correct via unnamed
        unnamed_arguments = list(orig_unnamed_arguments)
        unnamed_arguments.insert(3, add_inputs("a.test", "b.test"))
        results = parse_task_arguments(unnamed_arguments, {},
                                       expected_arguments, task_description)
        self.assertEqual(results, add_inputs_expected_results)

        # replace_inputs correct via named
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {"replace_inputs": ("a.test", "b.test")}, expected_arguments, task_description)
        self.assertEqual(results, replace_inputs_expected_results)

        # replace_inputs correct via named and paranoid inputs() wrapping
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {"replace_inputs": inputs(("a.test", "b.test"))}, expected_arguments, task_description)
        self.assertEqual(results, replace_inputs_expected_results)

        # replace_inputs correct via unnamed
        unnamed_arguments = list(orig_unnamed_arguments)
        unnamed_arguments.insert(3, inputs(("a.test", "b.test")))
        results = parse_task_arguments(unnamed_arguments, {},
                                       expected_arguments, task_description)
        self.assertEqual(results, replace_inputs_expected_results)

        # Error: both add_inputs and replace_inputs via named
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {"replace_inputs": ("a.test", "b.test"),
                                            "add_inputs": ("a.test", "b.test")}, expected_arguments, task_description)

        # Error: both add_inputs and replace_inputs via named / unnamed
        with self.assertRaises(error_too_many_args):
            unnamed_arguments = list(orig_unnamed_arguments)
            unnamed_arguments.insert(3, inputs(("a.test", "b.test")))
            results = parse_task_arguments(unnamed_arguments, {"add_inputs": ("a.test", "b.test")},
                                           expected_arguments, task_description)

        # Error: both add_inputs and replace_inputs via named / unnamed
        with self.assertRaises(error_too_many_args):
            unnamed_arguments = list(orig_unnamed_arguments)
            unnamed_arguments.insert(3, add_inputs("a.test", "b.test"))
            results = parse_task_arguments(unnamed_arguments, {"replace_inputs": ("a.test", "b.test")},
                                           expected_arguments, task_description)

        # Error: wrong number of arguments
        with self.assertRaises(error_inputs_multiple_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {"replace_inputs": inputs("a.test", "b.test")}, expected_arguments, task_description)

        # Error: no arguments
        with self.assertRaises(error_inputs_multiple_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {"replace_inputs": inputs()}, expected_arguments, task_description)


# ________________________________________________________________________________________________________

#   Test_parse_originate_args

# ________________________________________________________________________________________________________
class Test_parse_originate_args (unittest.TestCase):
    """
    Make sure @originate(output, extras...) works
    """

    def test_parse_originate_args(self):
        expected_arguments = ["output", "extras"]

        empty_unnamed_arguments = []
        empty_named_arguments = {}
        orig_unnamed_arguments = [["a.1", "b.1"], 1, 2, 3, 4]
        task_description = "@originate(%s)\ndef myfunc(...)\n"
        expected_results = {'output': orig_unnamed_arguments[0],
                            'extras': orig_unnamed_arguments[1:],
                            'named_extras': {},
                            }

        # Error: empty list
        with self.assertRaises(error_missing_args):
            parse_task_arguments(
                empty_named_arguments, empty_named_arguments, expected_arguments, task_description)

        # parse complete correctly
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {}, expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # All named parameters
        unnamed_arguments = orig_unnamed_arguments[0:2]
        results = parse_task_arguments([],
                                       {'output': orig_unnamed_arguments[0],
                                        'extras': orig_unnamed_arguments[1:]
                                        },
                                       expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # Error: Unknown named arguments
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments([],
                                           {'output': orig_unnamed_arguments[0],
                                            'what': orig_unnamed_arguments[1:]
                                            },
                                           expected_arguments, task_description)

        # Error: Duplicate named arguments
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {'output': orig_unnamed_arguments[0],
                                            },
                                           expected_arguments, task_description)

# _________________________________________________________________________________________

#   Test_parse_mkdir_args

# _________________________________________________________________________________________


class Test_parse_mkdir_args (unittest.TestCase):

    def test_parse_mkdir_args(self):
        expected_arguments = ["input", "filter", "output"]

        empty_unnamed_arguments = []
        empty_named_arguments = {}
        orig_unnamed_arguments = ["*.txt", suffix(".txt"), ".result"]
        task_description = "@mkdir(%s)\ndef myfunc(...)\n"
        expected_results = {'input': orig_unnamed_arguments[0],
                            'filter': orig_unnamed_arguments[1],
                            'output': orig_unnamed_arguments[2]}
        add_inputs_expected_results = {'input': orig_unnamed_arguments[0],
                                       'filter': orig_unnamed_arguments[1],
                                       'output': orig_unnamed_arguments[2]}
        replace_inputs_expected_results = {'input': orig_unnamed_arguments[0],
                                           'filter': orig_unnamed_arguments[1],
                                           'output': orig_unnamed_arguments[2]}

        # Error: empty list
        with self.assertRaises(error_missing_args):
            parse_task_arguments(
                empty_named_arguments, empty_named_arguments, expected_arguments, task_description)

        # parse complete correctly
        results = parse_task_arguments(orig_unnamed_arguments,
                                       {}, expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # Error: missing argument
        unnamed_arguments = orig_unnamed_arguments[0:1]
        with self.assertRaises(error_missing_args):
            results = parse_task_arguments(unnamed_arguments,
                                           {}, expected_arguments, task_description)

        # parse almost complete and rescued with named parameter
        unnamed_arguments = orig_unnamed_arguments[0:2]
        results = parse_task_arguments(unnamed_arguments,
                                       {'output': orig_unnamed_arguments[2]},
                                       expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # All named parameters
        unnamed_arguments = orig_unnamed_arguments[0:2]
        results = parse_task_arguments([],
                                       {'input': orig_unnamed_arguments[0],
                                        'filter': orig_unnamed_arguments[1],
                                        'output': orig_unnamed_arguments[2],
                                        },
                                       expected_arguments, task_description)
        self.assertEqual(results, expected_results)

        # Error: Unknown named arguments
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments([],
                                           {'input': orig_unnamed_arguments[0],
                                            'filter': orig_unnamed_arguments[1],
                                            'output': orig_unnamed_arguments[2],
                                            'what': orig_unnamed_arguments[3:]
                                            },
                                           expected_arguments, task_description)

        # Error: Duplicate named arguments
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments(orig_unnamed_arguments,
                                           {'input': orig_unnamed_arguments[0],
                                            },
                                           expected_arguments, task_description)

        # Error: no extras arguments allowed
        with self.assertRaises(error_too_many_args):
            results = parse_task_arguments(orig_unnamed_arguments + [1, formatter(), 'a', 4],
                                           {'input': orig_unnamed_arguments[0],
                                            },
                                           expected_arguments, task_description)


#
#   Necessary to protect the "entry point" of the program under windows.
#       see: http://docs.python.org/library/multiprocessing.html#multiprocessing-programming
#
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import unittest
import logging
from ruffus.proxy_logger import make_shared_logger_and_proxy, setup_std_shared_logger
from ruffus.cmdline import handle_verbose
from ruffus import *
import sys

"""

    test_cmdline.py



"""

import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name


#import traceback


class Test_Logging(unittest.TestCase):

    def test_rotating_log(self):
        """
            test rotating via proxy
        """
        open("/tmp/lg.log", "w").close()
        args = {}
        args["file_name"] = "/tmp/lg.log"
        args["rotating"] = True
        args["maxBytes"] = 20000
        args["backupCount"] = 10
        args["level"]= logging.INFO
        (my_log, logging_mutex) = make_shared_logger_and_proxy(
            setup_std_shared_logger, "my_logger", args)

        with logging_mutex:
            my_log.debug('This is a debug message')
            my_log.info('This is an info message')
            my_log.warning('This is a warning message')
            my_log.error('This is an error message')
            my_log.critical('This is a critical error message')
            my_log.log(logging.ERROR, 'This is a debug message')

        with open("/tmp/lg.log") as ii:
            data = ii.readlines()

            self.assertEqual(data,
                             ["This is an info message\n",
                              "This is a warning message\n",
                              "This is an error message\n",
                              "This is a critical error message\n",
                              "This is a debug message\n"])


if __name__ == '__main__':
    unittest.main()
